# coding: utf-8
"""This module provides a conversion layer for data types passed to/from
   Geoprocessing tasks on an ArcGIS REST server."""

import datetime
import json

from . import geometry

try:
    long, unicode, basestring
except NameError:
    long, unicode, basestring = int, str, str

class GPMultiValue(object):
    """Represents a multivalue Geoprocessing parameter"""
    _container_type = None
    def __init__(self, values):
        self._values = [self._container_type.fromJson(item)
                            if not isinstance(item, GPBaseType)
                            else item
                        for item in values]
    def __iter__(self):
        return iter(self._values)
    @classmethod
    def _from_json_def(cls, json):
        return cls
    @property
    def _json_struct(self):
        return [getattr(x, '_json_struct', x) for x in self._values]
    @classmethod
    def fromJson(cls, val):
        return cls(val)
    @staticmethod
    def fromType(datatype):
        """Use this method to create a MultiValue type from a specific GP Value
           type.
           
              >>> gpmultistr = arcrest.GPMultiValue.fromType(arcrest.GPString)
              >>> gpvalue = gpmultistr(["a", "b", "c"])
        """
        if issubclass(datatype, GPBaseType):
            return GPBaseType._get_type_by_name("GPMultiValue:%s" % 
                                                datatype.__name__)
        else:
            return GPBaseType._get_type_by_name("GPMultiValue:%s" % 
                                                str(datatype))

class GPBaseType(object):
    """Base type for (singular) Geoprocessing argument value types"""
    #: Mapping from Geoprocessing type to name
    _gp_type_mapping = {}

    def __str__(self):
        return json.dumps(self._json_struct)
    @classmethod
    def _from_json_def(cls, json):
        return cls
    @classmethod
    def _get_type_by_name(cls, name):
        if (name.startswith("GPMultiValue:") and 
                name not in cls._gp_type_mapping):
            def make_multivalue(name, base_type):
                new_multivalue_type = type(str(name), (GPMultiValue,), {})
                new_multivalue_type._container_type = base_type
                return new_multivalue_type
            mvs, dtype = name.split(":", 1)
            mytype = cls._gp_type_mapping.get(dtype, GPString)
            multivalue_type = make_multivalue(name, mytype)
            cls._gp_type_mapping[name] = multivalue_type
            return multivalue_type
        else:
            return cls._gp_type_mapping.get(name, GPString)
    @classmethod
    def _register_type(cls, newcls):
        cls._gp_type_mapping[newcls.__name__] = newcls
        return newcls

class GPSimpleType(GPBaseType):
    """For geoprocessing types that simplify to base Python types, such as 
       int, bool, or string."""
    __conversion__ = None
    def __init__(self, val):
        self.value = val
    @property
    def _json_struct(self):
        try:
            return self.__conversion__(self.value)
        except:
            if self.value is None:
                return None
            raise
    @classmethod
    def fromJson(cls, value):
        return cls.__conversion__(value)

@GPBaseType._register_type
class GPBoolean(GPSimpleType):
    """Represents a geoprocessing boolean parameter"""
    __conversion__ = bool

@GPBaseType._register_type
class GPDouble(GPSimpleType):
    """Represents a geoprocessing double parameter"""
    __conversion__ = float

@GPBaseType._register_type
class GPLong(GPSimpleType):
    """Represents a geoprocessing long parameter"""
    __conversion__ = long

@GPBaseType._register_type
class GPString(GPSimpleType):
    """Represents a geoprocessing string parameter"""
    __conversion__ = str

@GPBaseType._register_type
class GPLinearUnit(GPBaseType):
    """Represents a geoprocessing linear unit parameter"""
    #: The set of unit names allowed, defaulting to esriMeters
    allowed_units = set(['esriCentimeters',
                         'esriDecimalDegrees',
                         'esriDecimeters',
                         'esriFeet',
                         'esriInches',
                         'esriKilometers',
                         'esriMeters',
                         'esriMiles',
                         'esriMillimeters',
                         'esriNauticalMiles',
                         'esriPoints',
                         'esriUnknownUnits',
                         'esriYards'])
    def __init__(self, value, unit=None):
        if isinstance(value, (list, tuple)) and len(value) == 2:
            value, unit = value
        value = float(value)
        # Default to meters? Maybe take this out.
        if unit is None:
            unit = 'esriMeters'
        else:
            assert unit in self.allowed_units, "Unit %r not valid" % unit
        self.distance, self.units = value, unit
    @property
    def _json_struct(self):
        return {'distance': self.distance, 'units': self.units}
    @classmethod
    def fromJson(cls, val):
        return cls(val['distance'], val['units'])

def rowtuple(colnames):
    import operator
    class RowTuple(tuple):
        def __new__(cls, *values):
            return tuple.__new__(cls, *values)
    for i, col in enumerate(colnames):
        setattr(RowTuple, col, property(operator.itemgetter(i)))
    RowTuple.__slots__ = ()
    return RowTuple

@GPBaseType._register_type
class GPFeatureRecordSetLayer(GPBaseType):
    """Represents a geoprocessing feature recordset parameter"""
    _columns = None
    def __init__(self, Geometry, sr=None):
        if isinstance(Geometry, geometry.Geometry):
            Geometry = [Geometry]
        self._features = Geometry
        if sr:
            self.spatialReference = geometry.SpatialReference(sr)
        elif getattr(self, '_features', None) and len(self._features):
            self.spatialReference = geometry.SpatialReference(
                                        self._features[0].spatialReference)
        else:
            raise ValueError("Could not determine spatial reference")
        if self._columns is None and self._features:
            _columns = sorted(reduce(lambda x, y: x | y, 
                                   (set(getattr(row, 'attributes', {}).keys())
                                   for row in self._features)))
            if 'shape' not in (col.lower() for col in _columns):
                _columns = ['shape'] + _columns
            self._columns = tuple(_columns)
    @property
    def features(self):
        return list(self)
    def __iter__(self):
        return ({'geometry': feature, 
                 'attributes': getattr(feature, 'attributes', {})} 
                 for feature in self._features)
    @property
    def _json_struct(self):
        geometry_types = set(geom.__geometry_type__ for geom in self._features)
        assert len(geometry_types) == 1, "Must have consistent geometries"
        geometry_type = list(geometry_types)[0]
        return {
                    'geometryType': geometry_type,
                    'spatialReference': self.spatialReference._json_struct,
                    'features': [
                        x._json_struct_for_featureset for x in self._features]
               }
    @classmethod
    def fromJson(cls, value):
        spatialreference = geometry.fromJson(value['spatialReference']) \
            if 'spatialReference' in value else None
        geometries = [geometry.Polyline.fromCompressedGeometry(
                            geo['compressedGeometry'], geo['attributes'])
                      if "compressedGeometry" in geo
                      else geometry.fromJson(geo['geometry'], geo['attributes']) 
                      for geo in value['features']]
        return cls(geometries, spatialreference)

@GPBaseType._register_type
class GPRecordSet(GPBaseType):
    """Represents a geoprocessing recordset parameter"""
    _columns = None
    def __init__(self, arg, exceededTransferLimit=None):
        self.features = arg
        self._exceededTransferLimit = exceededTransferLimit
        if self._columns is None:
            _columns = sorted(reduce(lambda x, y: x | y, 
                                   (set(row['attributes'].keys()) 
                                   for row in self.features)))
            self._columns = tuple(_columns)
    def __iter__(self):
        return (feature for feature in self.features)
    @property
    def exceededTransferLimit(self):
        return self._exceededTransferLimit
    @classmethod
    def fromJson(cls, json):
        return cls(json['features'], json.get('exceededTransferLimit', None))
    @property
    def _json_struct(self):
        return { 'features': self.features }

@GPBaseType._register_type
class GPDate(GPBaseType):
    """Represents a geoprocessing date parameter. The format parameter
       in the object constructor varies from the REST API's format parameters
       in that each date field in the format string must be preceded by a
       percent sign, as in the strftime function. For further information
       about Python strftime format strings, please refer to
       http://docs.python.org/library/time.html#time.strftime"""
    #: default date format
    __date_format = "%a %b %d %H:%M:%S %Z %Y"
    #: secondary (fallback) date formats
    __secondary_date_formats = ["%c",
                                "%Y%m%dT%H:%M:%S",
                                "%Y-%m-%d %H:%M:%S",
                                "%m/%d/%Y %I:%M:%S %p"]
    def __init__(self, date, format=None):
        if isinstance(date, basestring):
            try:
                self.date = datetime.datetime.strptime(date, format)
            except (ValueError, TypeError) as e:
                try:
                    self.date = datetime.datetime.strptime(date, 
                                                           self.__date_format)
                    self.format = self.__date_format
                except ValueError:
                    for sformat in self.__secondary_date_formats:
                        try:
                            self.date = datetime.datetime.strptime(date, sformat)
                            self.format = format
                        except ValueError:
                            pass
                    if not getattr(self, 'date', None):
                        try:
                            import utils
                            self.date = utils.timetopythonvalue(date)
                        except:
                            pass
                        raise ValueError("Cannot convert %r to date" % date)
        elif isinstance(date, (datetime.date, datetime.datetime)):
            self.date = date
        else:
            try:
                import utils
                self.date = utils.timetopythonvalue(date)
            except:
                raise ValueError("Cannot convert %r to a date" % date)
        self.format = format
    @property
    def _json_struct(self):
        if self.format:
            return {'date': self.date.strftime(self.format),
                    'format': self.format.replace('%', '')}
        else:
            import utils
            return utils.pythonvaluetotime(self.date)
        #return self.date.strftime(self.__date_format)
    @classmethod
    def fromJson(cls, value):
        if isinstance(value, dict):
            datestring = value['date']
            formatstring = value['format']
            # Re-escape field names from formats like Y-m-d back to
            # strftime-style %Y-%m-%d strings
            for chr in "%aAbBcdHIjmMpSUwWxXyYZ":
                formatstring = formatstring.replace(chr, '%'+chr)
            return cls(datestring, formatstring)
        elif isinstance(value, basestring):
            return cls(value, cls.__date_format)

@GPBaseType._register_type
class GPDataFile(GPBaseType):
    """A URL for a geoprocessing data file parameter"""
    #: The URL of the data file
    url = ''
    def __init__(self, url):
        self.url = url
    @property
    def _json_struct(self):
        return {'url': self.url}
    @classmethod
    def fromJson(cls, value):
        return cls(value['url'])

class GPUrlWithFormatType(GPBaseType):
    """A class for representing Raster data and Raster layers -- both have a 
       URL and a Format attribute."""
    #: The URL of the resource
    url = ''
    #: The data format of the raster: jpeg, png, etc.
    format = ''
    def __init__(self, url, format):
        self.url, self.format = url, format
    @property
    def _json_struct(self):
        return {'url': self.url, 'format': self.format}
    @classmethod
    def fromJson(cls, value):
        return cls(value['url'], value['format'])

@GPBaseType._register_type
class GPRasterData(GPUrlWithFormatType):
    """A URL for a geoprocessing raster data file parameter, with format."""

@GPBaseType._register_type
class GPRasterDataLayer(GPUrlWithFormatType):
    """A URL for a geoprocessing raster data layer file parameter,
       with format."""

__all__ = sorted(GPBaseType._gp_type_mapping) + ['GPMultiValue']



# coding: utf-8
"""Represents the ArcGIS Portal REST APIs"""

from . import compat, server

__all__ = ['PortalRoot']

class PortalRoot(server.RestURL):
    def __init__(self, url, username=None, password=None,
                 token=None, generate_token=False,
                 expiration=60):
        url_list = list(compat.urlsplit(url))
        if not url_list[2].endswith('/'):
            url_list[2] += "/"
        url = compat.urlunsplit(url_list)
        if username is not None and password is not None:
            self._pwdmgr.add_password(None,
                                      url,
                                      username,
                                      password)
        if token:
            self.__token__ = token
        elif generate_token:
            self.__generateToken(url, username, password, expiration)
        super(PortalRoot, self).__init__(url)




# coding: utf-8
"""Implementation of the objects for the ArcGIS Server REST 
   Administration API"""

import cgi
import itertools
import os.path

from .. import compat, server, GenerateToken

__all__ = ['Admin', 'Folder', 'Services', 'Service', 
           'Machine', 'Machines', 'SiteMachines', 'ClusterMachines',
           'Directory', 'Directories',
           'Clusters', 'Cluster']

try:
    long, unicode, basestring
except NameError:
    long, unicode, basestring = int, str, str


class Admin(server.RestURL):
    """Represents the top level URL resource of the ArcGIS Server
       Administration API"""
    def __init__(self, url, username=None, password=None,
                 token=None, generate_token=False,
                 expiration=60):
        url_list = list(compat.urlsplit(url))
        if not url_list[2].endswith('/'):
            url_list[2] += "/"
        url = compat.urlunsplit(url_list)
        if username is not None and password is not None:
            self._pwdmgr.add_password(None,
                                      url,
                                      username,
                                      password)
        if token:
            self.__token__ = token
        elif generate_token:
            self.__generateToken(url, username, password, expiration)
        super(Admin, self).__init__(url)
    @property
    def resources(self):
        return self._json_struct['resources']
    @property
    def currentVersion(self):
        return self._json_struct['currentVersion']
    @property
    def clusters(self):
        return self._get_subfolder("./clusters/", Clusters)
    @property
    def services(self):
        return self._get_subfolder("./services/", Services)
    @property
    def machines(self):
        return self._get_subfolder("./machines/", SiteMachines)
    @property
    def data(self):
        return self._get_subfolder("./data/", Data)
    @property
    def uploads(self):
        return self._get_subfolder('./uploads/', Uploads)
    @property
    def system(self):
        return self._get_subfolder("./system/", System)
    def createNewSite(self, username, password, configStoreConnection=None, 
                      directories=None, cluster=None):
        res = self._get_subfolder("./createNewSite?f=json", 
                                  server.JsonPostResult,
                                  {'username': username,
                                   'password': password,
                                   'configStoreConnection': configStoreConnection,
                                   'directories': directories,
                                   'cluster': cluster})
        
        self.__generateToken(self.url, username, password, 60)
        return res
    def joinSite(self, adminURL, username, password):
        res = self._get_subfolder("./joinSite", 
                                  server.JsonPostResult,
                                  {'username': username,
                                   'password': password,
                                   'adminURL': adminURL})
        return res
    def deleteSite(self):
        res = self._get_subfolder("./deleteSite", 
                                  server.JsonPostResult)
        self.__token__ = None
        return res
    def __generateToken(self, url, username, password, expiration):
      token_auth = GenerateToken(url,
                                 username,
                                 password,
                                 expiration)
      if token_auth._json_struct.get('status', 'ok').lower() == 'error':
          raise compat.URLError('\n'.join(
                                      token_auth._json_struct.get(
                                          'messages', ['Failed.'])))
      self.__token__ = token_auth.token

class Data(server.RestURL):
    """Administration URL's data store -- Geodatabases and file data"""
    @property
    def geodatabases(self):
        return self._get_subfolder("./geodatabases/", GeoDatabases)
    @property
    def items(self):
        return self._get_subfolder("./items/", DataItems)

class GeoDatabases(server.RestURL):
    """Server's geodatabases and GDB connections"""
    pass

class HasUploads(object):
    def upload(self, file, description=''):
        if isinstance(file, basestring):
            file = open(file, 'rb')
        sub = self._get_subfolder('./upload/', server.JsonResult,
                                  {'description': description},
                                  {'itemFile': file})
        return sub._json_struct['item']

class Uploads(server.RestURL, HasUploads):
    """Uploads URL"""
    pass

class DataItems(server.RestURL, HasUploads):
    """Server's data files"""
    @property
    def packages(self):
        return self._json_struct['packages']

class Folder(server.RestURL):
    @property
    def folderName(self):
        return self._json_struct['folderName']
    @property
    def description(self):
        return self._json_struct['description']
    @property
    def serviceNames(self):
        return [service['serviceName'] 
                for service in self._json_struct['services']]
    @property
    def services(self):
        return [self._get_subfolder("./%s.%s/" % 
                                        (servicename['serviceName'],
                                         servicename['type']), 
                                    Service)
                for servicename in self._json_struct['services']]
    def __getitem__(self, itemname):
        if '/' in itemname:
            itemname, rest = itemname.split('/', 1)
            return self[itemname][rest]
        for servicename in self._json_struct['services']:
            fstrings = (servicename['serviceName'].lower(),
                        (servicename['serviceName'] +
                            "." + 
                            servicename['type']).lower())
            if itemname.lower() in fstrings:
                return self._get_subfolder("./%s.%s/" % 
                                        (servicename['serviceName'],
                                         servicename['type']), 
                                    Service)
        raise KeyError(itemname)
    def __iter__(self):
        return iter(self.services)

class Services(Folder):
    def createFolder(self, folderName, description):
        raise NotImplementedError("Not implemented")
    @property
    def folders(self):
        return [self._get_subfolder("./%s/" % foldername, Folder) 
                for foldername in self._json_struct['folders']
                if foldername != "/"]
    @property
    def types(self):
        return_type = self._get_subfolder("./types/", server.JsonPostResult)
        return return_type._json_struct['types']
    def __getitem__(self, itemname):
        for foldername in self._json_struct['folders']:
            if foldername.lower() == itemname.lower():
                return self._get_subfolder("./%s/" % foldername, Folder)
        return super(Services, self).__getitem__(itemname)
    def __iter__(self):
        for folder in self.folders:
            for service in folder.services:
                yield service
        for service in super(Services, self).__iter__():
            yield service

class Service(server.RestURL):
    @property
    def name(self):
        return self._json_struct['serviceName'] + "." + self._json_struct['type']
    @property
    def status(self):
        return self._get_subfolder("./status/", 
                                   server.JsonPostResult)._json_struct
    @property
    def statistics(self):
        return self._get_subfolder("./statistics/",
                                   server.JsonPostResult)._json_struct
    def start(self):
        return self._get_subfolder("./start/", 
                                   server.JsonPostResult)._json_struct
    def stop(self):
        return self._get_subfolder("./stop/", 
                                   server.JsonPostResult)._json_struct
    def delete(self):
        return self._get_subfolder("./delete/", 
                                   server.JsonPostResult)._json_struct

class Machine(server.RestURL):
    """Base class for a single machine on a site"""
    @property
    def name(self):
        return self._json_struct['machineName']
    @property
    def admin_url(self):
        return self._json_struct['adminURL']
    @property
    def platform(self):
        return self._json_struct['platform']
    def start(self):
        return self._get_subfolder("./start/", 
                                   server.JsonPostResult)._json_struct
    def stop(self):
        return self._get_subfolder("./stop/", 
                                   server.JsonPostResult)._json_struct
    def unregister(self):
        return self._get_subfolder("./unregister/", 
                                   server.JsonPostResult)._json_struct

class Machines(server.RestURL):
    """Base class for a list of machines, both on a Cluster and a Site"""
    __post__ = True
    __machines__ = Ellipsis
    @property
    def _machines(self):
        if self.__machines__ is Ellipsis:
            path_and_attribs = [(d['machineName'], d) 
                                for d in self._json_struct['machines']]
            self.__machines__ = dict(path_and_attribs)
        return self.__machines__
    def keys(self):
        return self._machines.keys()
    def __iter__(self):
        return (Admin(item['adminURL']) 
                    for item in self._machines.itervalues())
    def register(self, machine_name, admin_url=None):
        return self._get_subfolder("./register/", 
                                   server.JsonPostResult,
                                   {'machineName': machine_name,
                                    'adminURL': admin_url})._json_struct


class ClusterMachines(Machines):
    """A list of machines participating on a cluster"""
    def add(self, machine_names):
        if isinstance(machine_names, basestring):
            machine_names = [machine_names]
        responses = [self._get_subfolder("./add/", server.JsonPostResult, 
                                         {"machineNames": m}) 
                     for m in machine_names]
        return responses
    def remove(self, machine_names):
        if isinstance(machine_names, basestring):
            machine_names = [machine_names]
        responses = [self._get_subfolder("./remove/", server.JsonPostResult, 
                                         {"machineNames": m}) 
                     for m in machine_names]
        return responses

class SiteMachines(Machines):
    """A list of machines on a site"""
    def register(self, machineName, adminURL=None):
        res = self._get_subfolder("./register/", server.JsonPostResult,
                                  {'machineName': machineName,
                                   'adminURL': adminURL})
    @property
    def machines(self):
        return [self._get_subfolder("./%s/" % machinename, Machine) for
                machinename in self._machines]
    def __getitem__(self, itemname):
        assert itemname in self._machines, "Couldn't find %s" % itemname
        return self._get_subfolder('./%s/' % itemname, Machine)

class Directory(server.RestURL):
   __post__ = True

class Directories(server.RestURL):
    __directories__ = Ellipsis
    @property
    def _directories(self):
        path_and_attribs = [(d['physicalPath'], d) 
                            for d in self._json_struct['directories']]
        self.__directories__ = dict(path_and_attribs)
        return self.__directories__
    def __contains__(self, k):
        return self._directories.__contains__(k)
    def __getitem__(self, k):
        return self._directories.__getitem__(k)
    def register(self, type, path, vpath=None):
        response = self._get_subfolder('./register', server.JsonPostResult,
                                      {'directoryType': type.upper(),
                                       'physicalPath': path,
                                       'virtualPath': vpath})._json_struct
    def unregister(self, path):
        response = self._get_subfolder('./unregister', server.JsonPostResult,
                                      {'physicalPath': path})._json_struct

class Cluster(server.JsonResult):
    __post__ = True
    __lazy_fetch__ = False
    __cache_request__ = True
    def __eq__(self, other):
        if not isinstance(other, Cluster):
            return False
        return self._url == other._url
    @property
    def machineNames(self, _error=None, _success=None):
        if "machineNames" in self._json_struct:
            return self._json_struct["machineNames"]
    @property
    def machines(self):
        return self._get_subfolder("./machines/", ClusterMachines)
    def start(self):
        return self._get_subfolder('./start/',
                                   server.JsonPostResult)._json_struct
    def stop(self):
        return self._get_subfolder('./stop/',
                                   server.JsonPostResult)._json_struct
    def delete(self):
        return self._get_subfolder('./delete/',
                                   server.JsonPostResult)._json_struct
    def editProtocol(self, type="TCP", tcpClusterPort=-1, 
               multicastAddress=10, multicastPort=-1):
        if type not in ("TCP", "UDP"):
            raise ValueError("Got %r. Valid choices are: TCP, UDP" % type)
        res = self._get_subfolder('./editProtocol', server.JsonPostResult,
                                     {'type': type,
                                      'tcpClusterPort': tcpClusterPort 
                                                            if type == "TCP" 
                                                            else None,
                                      'multicastAddress': multicastAddress 
                                                            if type == "UDP" 
                                                            else None,
                                      'multicastPort': multicastPort
                                                            if type == "UDP" 
                                                            else None})

class Clusters(server.RestURL):
    __post__ = True
    __directories__ = Ellipsis
    __cluster_cache__ = Ellipsis
    @property
    def _clusters(self):
        if self.__cluster_cache__ is Ellipsis:
            path_and_attribs = [(d['clusterName'],
                                self._get_subfolder('./%s/' %d['clusterName'],
                                                    Cluster)) 
                                for d in self._json_struct['clusters']] 
            self.__cluster_cache__ = dict(path_and_attribs)
        return self.__cluster_cache__
    @property
    def clusterNames(self):
        return [d['clusterName'] for d in self._json_struct['clusters']]
    def __contains__(self, k):
        if isinstance(k, int):
            return k < len(self)
        return self._clusters.__contains__(k)
    def __getitem__(self, k):
        if isinstance(k, int):
            k = self.clusterNames[k]
        return self._clusters.__getitem__(k)
    def __len__(self):
        return len(self.clusterNames)
    def create(self, clusterName, type="TCP", tcpClusterPort=-1, 
               multicastAddress=10, multicastPort=-1):
        if type not in ("TCP", "UDP"):
            raise ValueError("Got %r. Valid choices are: TCP, UDP" % type)
        res = self._get_subfolder('./create', server.JsonPostResult,
                                     {'clusterName': clusterName,
                                      'type': type,
                                      'tcpClusterPort': tcpClusterPort 
                                                            if type == "TCP" 
                                                            else None,
                                      'multicastAddress': multicastAddress 
                                                            if type == "UDP" 
                                                            else None,
                                      'multicastPort': multicastPort
                                                            if type == "UDP" 
                                                            else None})
        if 'error' in res._json_struct:
            raise Exception(repr(res._json_struct))
        return self._get_subfolder('./%s/' % clusterName, Cluster)




# coding: utf-8
"""Arcrest is a Python binding to the ArcGIS REST Server API, similar to the
   JavaScript or Flex API in program structure as well as in the way it
   interfaces with ArcGIS servers.
   
   Getting Started with Arcrest
   ============================

   A simple example of connecting to a server:

      >>> import arcrest
      >>> catalog = arcrest.Catalog("http://sampleserver1.arcgisonline.com/arcgis/rest/services")
      >>> catalog.services
      [<GeometryServer ('http://sampleserver1.arcgisonline.com/arcgis/rest/services/Geometry/GeometryServer/?f=json')>]

   Getting a service from a catalog:

      >>> locator = catalog.Locators.ESRI_Geocode_USA
      >>> locator.url
      'http://sampleserver1.arcgisonline.com/arcgis/rest/services/Locators/ESRI_Geocode_USA/GeocodeServer/?f=json'
      >>> locator.FindAddressCandidates(Address="380 NEW YORK ST", City="Redlands", State="CA")
      <FindAddressCandidatesResult('http://sampleserver1.arcgisonline.com/arcgis/rest/services/Locators/ESRI_Geocode_USA/GeocodeServe...')>
      >>> candidates.candidates[0]
      {'attributes': {}, 'score': 81, 'location': POINT(-117.19568 34.05752), 'address': '380 NEW YORK ST, REDLANDS, CA, 92373'}

   Getting a service from a URL:

      >>> service = arcrest.GPService("http://sampleserver1.arcgisonline.com/ArcGIS/rest/services/Specialty/ESRI_Currents_World/GPServer")
      >>> service.url
      'http://sampleserver1.arcgisonline.com/ArcGIS/rest/services/Specialty/ESRI_Currents_World/GPServer/?f=json'

   Inspecting and executing a geoprocessing service:

      >>> service.tasks
      [<GPTask('http://sampleserver1.arcgisonline.com/ArcGIS/rest/services/Specialty/ESRI_Currents_World/GPServer...')>]
   
   Getting a task by name:

      >>> task = service.MessageInABottle
      >>> task.url
      'http://sampleserver1.arcgisonline.com/ArcGIS/rest/services/Specialty/ESRI_Currents_World/GPServer/MessageInABottle/?f=json'      
      >>> task.synchronous
      True
      >>> task.name
      'MessageInABottle'

   Inspecting tasks:

      >>> [(param['name'], param['dataType']) for param in task.parameters if param['direction'] == 'esriGPParameterDirectionInput']
      [('Input_Point', 'GPFeatureRecordSetLayer'), ('Days', 'GPDouble')]

   Executing a job:

      >>> results = service.MessageInABottle(arcrest.Point(-11, 38, arcrest.projections.geographic.GCS_WGS_1984), 2)
      >>> import time
      >>> while results.running:
      ...   time.sleep(0.25)
      ...      
      >>> results.Output.features
      [{'geometry': MULTILINESTRING((-11.00000 38.00000,-10.85327 37.73851,-10.83942 37.71683,-10.83734 37.71359,-10.83702 37.71310,-10.83697 37.71303)), 'attributes': {'shape_length': 0.330091748127675, 'fid': 1, 'fnode_': 0}}]

   """

from arcrest.geometry import *
from arcrest.gptypes import *
from arcrest.server import *
from arcrest.projections import projected, geographic



# coding: utf-8
"""Provides a place for functions/modules which have been reogranized in the
   python 2/3 switch use in this library to be located regardless of their
   location in the running Python's standard library."""

__all__ = ['cookielib', 'urllib2', 'HTTPError', 'URLError', 'urlsplit',
           'urljoin', 'urlunsplit', 'urlencode', 'quote', 'string_type',
           'ensure_string', 'ensure_bytes', 'get_headers']

try:
    long, unicode, basestring
except NameError:
    long, unicode, basestring = int, str, str

try:
    import cookielib
except ImportError:
    import http.cookiejar as cookielib

try:
    import urllib2
except ImportError:
    import urllib.request as urllib2

try:
    from urllib2 import HTTPError, URLError
except ImportError:
    from urllib.error import HTTPError, URLError

try:
    from urlparse import urlsplit, urljoin, urlunsplit
except ImportError:
    from urllib.parse import urlsplit, urljoin, urlunsplit

try:
    from urllib import urlencode, quote
except ImportError:
    from urllib.parse import urlencode, quote

string_type = str

try:
    unicode
    string_type = basestring
    bytes = str
except NameError:
    unicode = str
    string_type = str

def ensure_string(payload_bytes):
    if isinstance(payload_bytes, bytes):
        return payload_bytes.decode("utf-8")
    return payload_bytes

def ensure_bytes(payload_string):
    if isinstance(payload_string, unicode):
        return payload_string.encode("utf-8")
    return payload_string

def get_headers(handle):
    if hasattr(handle.headers, 'headers'):
        return handle.headers.headers
    return dict(handle.headers.items())



# coding: utf-8
from __future__ import print_function

import argparse
import json
import os
import sys
import time

from .. import compat, Catalog

__all__ = ['createservice', 'manageservice', 'managesite', 'deletecache',
           'managecachetiles', 'createcacheschema',
           'convertcachestorageformat', 'importcache', 'exportcache']

IN_WINE = any(e in os.environ for e in ('WINELOADERNOEXEC',
                                        'DOTWINE',
                                        'WINEPREFIX',
                                        'WINEDEBUG'))
PROG_NAME = os.path.basename((sys.argv[0] 
                              if not IN_WINE
                              else os.path.splitext(sys.argv[0])[0]))

shared_args = argparse.ArgumentParser(prog=PROG_NAME, add_help=False)
shared_args.add_argument('-u', '--username', 
                         required=True,
                         help='Description: Username for Server')
shared_args.add_argument('-p', '--password', 
                         required=True,
                         help='Description: Password for Server')
shared_args.add_argument('-s', '--site', 
                         required=True,
                         help='Description: URL for admin Server, '
                              'e.g. http://host:6080/arcgis')

class ActionNarrator(object):
    def __init__(self):
        self.action_stack = []
    def __call__(self, action):
        self.action = action
        return self
    def __enter__(self):
        self.action_stack.append(self.action)
    def __exit__(self, t, ex, tb):
        action = self.action_stack.pop()
        if (t, ex, tb) != (None, None, None):
            #import traceback
            #traceback.print_exception(t, ex, tb)
            if t is not SystemExit:
                print("Error {0}: {1}".format(action, str(ex)))
            sys.exit(1)
        
def get_rest_urls(server_url):
    
    if not server_url.endswith('/'):
         server_url += '/'
         
    urllist = compat.urlsplit(server_url)
    d = urllist._asdict()
                    
    context = d['path']
    admin_url = server_url 
    if (context != '/'):
        if (not context.endswith('admin/')):
            admin_url = compat.urljoin(server_url, context + 'admin/')
        else:
            admin_url = compat.urljoin(server_url, context)
        if (not context.endswith('rest/services/')):
            rest_url = compat.urljoin(server_url, context+'rest/services/')
        else:
            rest_url = compat.urljoin(server_url, context)
    else:
        admin_url = compat.urljoin(server_url, 'arcgis/admin/')
        rest_url = compat.urljoin(server_url, 'arcgis/rest/services/')
    return (admin_url, rest_url)

def provide_narration(fn):
    def fn_():
        return fn(ActionNarrator())
    return fn_

def wait_on_tool_run(result_object, silent=False):
    message_length = 0
    while result_object.running:
        time.sleep(0.125)
        messages = getattr(result_object, 'messages', [])
        if len(messages) > message_length:
            for message_object in messages[message_length:]:
                if not silent:
                    print (message_object.description)
                    sys.stdout.flush()
            message_length = len(messages)
    if not silent:
        for message_object in messages[message_length:]:
            print (message_object.description)

createserviceargs = argparse.ArgumentParser(prog=PROG_NAME,
                                            description='Creates a service',
                                            parents=[shared_args])
createserviceargs.add_argument('-C', '--cluster',
                               nargs='?',
                               default=None,
                               help='Name of cluster to act on')
createserviceargs.add_argument('-f', '--sdfile',
                                nargs='+',
                                required=True,
                                help='Filename of local Service Definition '
                                     'file')
createserviceargs.add_argument('-F', '--folder-name',
                               nargs='?',
                               default=None,
                               help='Folder to create service in')
createserviceargs.add_argument('-n', '--service-name',
                               nargs='?',
                               default=None,
                               help='Name of service to create')
createserviceargs._optionals.title = "arguments"

@provide_narration
def createservice(action):
    import arcrest.admin as admin
    args = createserviceargs.parse_args()
    files = args.sdfile
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to admin site {0}".format(admin_url)):
        site = admin.Admin(admin_url, args.username, args.password,
                           generate_token=True)
        assert site._json_struct.get('status', 'ok') != 'error',\
               ' '.join(site._json_struct.get('messages',
                   ['Could not connect to site.']))
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("looking up Publish Tool"):
        publish_tool = (rest_site['System']
                                 ['PublishingTools']
                                 ['Publish Service Definition'])
    with action("looking up cluster"):
        cluster = site.clusters[args.cluster] if args.cluster else None
    with action("verifying service definition file exists"):
        all_files = [os.path.abspath(filename) for filename in files]
        assert all_files, "No file specified"
        for filename in all_files:
            assert os.path.exists(filename) and os.path.isfile(filename), \
                    "{0} is not a file".format(filename)
    ids = []
    publish_tool.__post__ = True
    for filename in all_files:
        with action("uploading and publishing {0}".format(
                                                  os.path.basename(filename))):
            id = site.uploads.upload(filename)['itemID']
            config_url = compat.urljoin(site.uploads.url, 
                                     '{}/serviceconfiguration.json'.format(id))
            with action("fetching default configuration"):
                config_url += "?token={}".format(site.__token__)
                config_json = json.load(compat.urllib2.urlopen(config_url))
            with action("adjusting service configuration with user options"):
                if args.folder_name and 'folderName' in config_json:
                    config_json['folderName'] = args.folder_name
                if args.service_name and 'service' in config_json \
                        and 'serviceName' in config_json['service']:
                    config_json['service']['serviceName'] = args.service_name
            with action("publishing {0}".format(os.path.basename(filename))):
                new_json = json.dumps(config_json)
                result_object = publish_tool(id, new_json, "")
                wait_on_tool_run(result_object, silent=True)

manageserviceargs = argparse.ArgumentParser(prog=PROG_NAME, description=
                                                'Manages/modifies a service',
                                            parents=[shared_args])
manageserviceargs.add_argument('-n', '--name',
                               default=None,
                               help='Description: Service name (optional)')
manageserviceargs.add_argument('-o', '--operation',
                               default=None,
                               help="Description: Operation to perform on "
                                    "specified service. If -l or --list is "
                                    "specified, used as a status filter "
                                    "instead.",
                               choices=['status', 'start', 'stop', 'delete'])
manageserviceargs.add_argument('-l', '--list',
                               default=False,
                               action='store_true',
                               help="Description: List services on server "
                                    "(optional)")
manageserviceargs._optionals.title = "arguments"

@provide_narration
def manageservice(action):
    import arcrest.admin as admin
    args = manageserviceargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    if args.list:
        with action("checking arguments"):
            assert not args.name, "name cannot be set if listing services"
        with action("connecting to admin site {0}".format(admin_url)):
            site = admin.Admin(admin_url, args.username, args.password,
                                          generate_token=True)
            assert site._json_struct.get('status', 'ok') != 'error',\
                   ' '.join(site._json_struct.get('messages',
                       ['Could not connect to site.']))
        with action("listing services"):
            services = site.services
            folders = services.folders
        with action("printing services"):
            status_map = {'stopped': 'stop',
                          'started': 'started'}
            servicelist = list(services.services)
            for folder in folders:
                servicelist += list(folder.services)
            for service in servicelist:
                print("{0:40} | {1}".format(
                                (service.parent.folderName+"/"
                                    if service.parent.folderName != "/"
                                    else ""
                                )+service.name,
                                service.status['realTimeState']))
    else:
        with action("checking arguments"):
            assert args.name, "Service name not specified"
        with action("connecting to admin site {0}".format(admin_url)):
            site = admin.Admin(admin_url, args.username, args.password,
                               generate_token=True)
            assert site._json_struct.get('status', 'ok') != 'error',\
                   ' '.join(site._json_struct.get('messages',
                       ['Could not connect to site.']))
        with action("listing services"):
            services = site.services
        with action("searching for service %s" % args.name):
            service = services[args.name]
        operation = (args.operation or '').lower()
        if operation == 'status':
            for key, item in sorted(service.status.items()):
                print("{0}: {1}".format(key, item))
        elif operation == 'start':
            with action("starting service"):
                return service.start()
        elif operation == 'stop':
            with action("stopping service"):
                return service.stop()
        elif operation == 'delete':
            with action("deleting service"):
                return service.delete()

managesiteargs = argparse.ArgumentParser(prog=PROG_NAME,
                                         description=
                                                'Manages/modifies a site',
                                            parents=[shared_args])
managesiteargs.add_argument('-A', '--add-machines',
                               nargs='+',
                               help='Machines to add to cluster')
managesiteargs.add_argument('-R', '--remove-machines',
                               nargs='+',
                               help='Machines to remove from cluster')
managesiteargs.add_argument('-l', '--list',
                               default=False,
                               action='store_true',
                               help='List machines on a site')
managesiteargs.add_argument('-lc', '--list-clusters',
                               default=False,
                               action='store_true',
                               help='List clusters on a site')
managesiteargs.add_argument('-o', '--operation',
                               nargs='?',
                               help='Description: Operation to perform on '
                                    'cluster',
                               choices=['chkstatus', 'start', 'stop'])
managesiteargs.add_argument('-c', '--cluster',
                               nargs='?',
                               default=None,
                               help='Name of cluster to act on')
managesiteargs.add_argument('-D', '--delete-cluster',
                               default=False,
                               action='store_true',
                               help='Delete cluster specified with -c')
managesiteargs.add_argument('-cr', '--create-cluster',
                               default=False,
                               action='store_true',
                               help=('Create cluster specified with -c '
                                     'if it does not exist'))
managesiteargs.add_argument('-cp', '--cluster-port',
                               type=int, nargs=1, default=4004, required=False,
                               help=('Port to use for cluster (if '
                                     'creating new)'))
managesiteargs._optionals.title = "arguments"

@provide_narration
def managesite(action):
    import arcrest.admin as admin
    args = managesiteargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to admin site {0}".format(admin_url)):
        site = admin.Admin(admin_url, args.username, args.password,
                           generate_token=True)
        assert site._json_struct.get('status', 'ok') != 'error',\
               ' '.join(site._json_struct.get('messages',
                   ['Could not connect to site.']))
    with action("determining actions to perform"):
        assert any([
                     args.add_machines, 
                     args.remove_machines,
                     args.delete_cluster,
                     args.create_cluster,
                     args.list,
                     args.list_clusters,
                     args.operation
                ]), "No action specified (use --help for options)"
    operation = (args.operation or '').lower()
    if not args.list_clusters:
        with action("looking up cluster"):
            try:
                cluster = site.clusters[args.cluster] if args.cluster else None
            except KeyError:
                if args.create_cluster:
                    cluster = site.clusters.create(args.cluster,
                                                   tcpClusterPort=
                                                        args.cluster_port)
                else:
                    raise
        with action("performing {0}".format(operation or '')):
            assert cluster, "No cluster specified"
            if operation.lower() == "start":
                cluster.start()
            elif operation.lower() == "stop":
                cluster.stop()
            elif operation == "chkstatus":
                raise NotImplementedError("Chkstatus not implemented")
        with action("deleting cluster"):
            if args.delete_cluster:
                assert cluster, "No cluster specified to delete"
                cluster.delete()
        with action("adding machines to cluster"):
            if args.add_machines:
                assert cluster, "No cluster specified"
                for machine in args.add_machines:
                    with action("adding {0} to cluster".format(machine)):
                        cluster.machines.add(machine)
        with action("removing machines from cluster"):
            if args.remove_machines:
                assert cluster, "No cluster specified"
                for machine in args.remove_machines:
                    with action("deleting {0} from cluster".format(machine)):
                        cluster.machines.remove(machine)
        with action("listing machines"):
            if args.list:
                name, itemobject = ('cluster', cluster) \
                        if cluster else ('site', site)
                print("===Machines on this {0}===".format(name))
                for machine in itemobject.machines.keys():
                    print("-", machine)
                print()
    elif args.list_clusters:
        with action("listing clusters"):
            if args.list_clusters:
                print("===Clusters on this site===")
                for cluster in site.clusters.clusterNames:
                    print("-", cluster)
                print()

deletecacheargs = argparse.ArgumentParser(prog=PROG_NAME,
                                          description=
                                                'Deletes a map tile cache',
                                            parents=[shared_args])
deletecacheargs.add_argument('-n', '--name',
                               help='Description: Service name')
deletecacheargs.add_argument('-i', '--instances',
                             help="Number of instances",
                             default=3,
                             type=int)
deletecacheargs._optionals.title = "arguments"

@provide_narration
def deletecache(action):
    import arcrest.admin as admin
    args = deletecacheargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("fetching reference to Delete Cache tool"):
        delete_cache_tool = (rest_site['System']
                                      ['CachingTools']
                                      ['DeleteCache'])
    with action("deleting map cache"):
        result_object = delete_cache_tool(args.name, args.instances)
        wait_on_tool_run(result_object, silent=True)

managecachetilesargs = argparse.ArgumentParser(prog=PROG_NAME,
                                               description=
                                                'Manage a map tile cache',
                                               parents=[shared_args])
managecachetilesargs.add_argument('-n', '--name',
                               help='Description: Service name')
managecachetilesargs.add_argument('-scales',
                                  help=
                                   "Description: Scales to generate caches")
managecachetilesargs.add_argument('-mode', '--update-mode' ,
                                  help="Description: Update mode",
                                  choices=['RECREATE_ALL_TILES',
                                           'RECREATE_EMPTY_TILES',
                                           'DELETE_TILES'])
managecachetilesargs.add_argument('-i', '--instances',
                                  help="Number of instances",
                                  default=3,
                                  type=int)
managecachetilesargs.add_argument('-AOI', '--feature-class',
                                  default=None,
                                  help="Description: Feature class for "
                                       "area of interest")
managecachetilesargs.add_argument('-extent', '--cache-extent',
                                  metavar='"{xmin; ymin; xmax; ymax}"',
                                  help="Extent[s] to cache",
                                  nargs='+')
managecachetilesargs.add_argument('-wait', '--wait-for-completion',
                            help="Description: Wait for operation to complete",
                                  choices=['WAIT', 'DO_NOT_WAIT'],
                                  default="WAIT")
managecachetilesargs._optionals.title = "arguments"

@provide_narration
def managecachetiles(action):
    import arcrest.admin as admin
    args = managecachetilesargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("fetching reference to Manage Cache tool"):
        manage_cache_tool = (rest_site['System']
                                      ['CachingTools']
                                      ['Manage Map Cache Tiles'])
    with action("managing map cache"):
        result_object = manage_cache_tool(args.name,
                                          args.scales,
                                          args.instances,
                                          args.update_mode,
                                          args.cache_extent,
                                          args.feature_class)
    if args.wait_for_completion.lower() == "wait":
        wait_on_tool_run(result_object)
    else:
        print (result_object.url)

createcacheschemaargs = argparse.ArgumentParser(prog=PROG_NAME,
                                                description=
                                             'Creates a map tile cache schema',
                                                parents=[shared_args])
createcacheschemaargs.add_argument('-n', '--name',
                               help='Description: Service name '
                                    '(format as ServiceName:ServiceType)')
createcacheschemaargs.add_argument('-Dc', '--cache_directory',
                               help='Description: ArcGIS Server Cache '
                                    'Directory')
createcacheschemaargs.add_argument('-scales', '--custom-scale-values',
                               help='Description: Scales (if Custom)',
                               default="",
                               nargs='+')
createcacheschemaargs.add_argument('-dpi', '--DPI',
                               help='Description: DPI of tiles [0-100]',
                               type=int,
                               default=100,
                               metavar='0-100')
createcacheschemaargs.add_argument('-TS', '--tile-size',
                               help='Description: Tile size',
                               choices=['125x125',
                                        '256x256',
                                        '512x512',
                                        '1024x1024'],
                               default='125x125')
createcacheschemaargs.add_argument('-TO', '--tile-origin',
                               help='Description: Tile origin. '
                                    'Ex: "-20037508.342787 20037508.342787"')
createcacheschemaargs.add_argument('-TF', '--tile-format',
                               help='Description: Tile format',
                               choices=['PNG', 'PNG8', 'PNG24', 'PNG32',
                                        'JPEG', 'MIXED'])
createcacheschemaargs.add_argument('-TC', '--tile-compression',
                               help='Description: Compression (if JPEG or '
                                    'MIXED)',
                               default=0,
                               type=int,
                               metavar='0-100')
createcacheschemaargs.add_argument('-storage', '--tile-storage-format',
                               help='Description: Tile storage format',
                               choices=['COMPACT', 'EXPLODED'])

createcacheschemaargs._optionals.title = "arguments"

@provide_narration
def createcacheschema(action):
    import arcrest.admin as admin
    args = createcacheschemaargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("fetching reference to Import Cache tool"):
        create_cache_tool = (rest_site['System']
                                      ['CachingTools']
                                      ['Create Map Cache'])
    with action("creating map cache"):
        result_object = create_cache_tool(args.name,
                                          args.cache_directory,
                                          args.tile_origin,
                                          args.custom_scale_values,
                                          args.tile_storage_format,
                                          args.tile_format,
                                          args.tile_compression,
                                          args.DPI,
                                          args.tile_size.split("x")[0],
                                          args.tile_size.split("x")[1])
        wait_on_tool_run(result_object)

convertcachestorageformatargs = argparse.ArgumentParser(prog=PROG_NAME,
                                            description=
                                             'Convert the format of a stored '
                                             'map cache',
                                            parents=[shared_args])

convertcachestorageformatargs.add_argument('-n', '--name',
                                  help="Service name")
convertcachestorageformatargs.add_argument('-i', '--instances',
                                  help="Number of instances",
                                  default=3,
                                  type=int)

convertcachestorageformatargs._optionals.title = "arguments"

@provide_narration
def convertcachestorageformat(action):
    import arcrest.admin as admin
    args = convertcachestorageformatargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("fetching reference to Convert Cache Storage Format tool"):
        convert_cache_tool = (rest_site['System']
                                       ['CachingTools']
                                       ['Convert Cache Storage Format'])
    with action("converting format"):
        result_object = convert_cache_tool(args.name,
                                           args.instances)
        while result_object.running:
            time.sleep(0.125)
        print ("\n".join(msg.description for msg in result_object.messages))

importcacheargs = argparse.ArgumentParser(prog=PROG_NAME,
                                          description=
                                             'Import a stored map cache',
                                            parents=[shared_args])

importcacheargs.add_argument('-n', '--name',
                                  help="Service name")
importcacheargs.add_argument('-i', '--instances',
                                  help="Number of instances",
                                  type=int)
importcacheargs.add_argument('-dC', '--cache-directory',
                             help="Cache directory")

@provide_narration
def importcache(action):
    import arcrest.admin as admin
    args = importcacheargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("fetching reference to Import Cache tool"):
        convert_cache_tool = (rest_site['System']
                                       ['CachingTools']
                                       ['Import Cache'])
    with action("importing cache"):
        result_object = convert_cache_tool(args.site[:args.site.find('?')]
                                            if '?' in args.site
                                            else args.site,
                                           args.target_cache,
                                           args.instances,
                                           args.extent,
                                           args.levels)
        while result_object.running:
            time.sleep(0.125)
        print ("\n".join(msg.description for msg in result_object.messages))

@provide_narration
def exportcache(action):
    import arcrest.admin as admin
    args = exportcacheargs.parse_args()
    admin_url, rest_url = get_rest_urls(args.site)
    with action("connecting to REST services {0}".format(rest_url)):
        rest_site = Catalog(rest_url, args.username, args.password,
                            generate_token=True)
    with action("fetching reference to Convert Cache Storage Format tool"):
        convert_cache_tool = (rest_site['System']
                                       ['CachingTools']
                                       ['Convert Cache Storage Format'])
    with action("converting format"):
        result_object = convert_cache_tool(args.name,
                                           args.instances)
        while result_object.running:
            time.sleep(0.125)
        print ("\n".join(msg.description for msg in result_object.messages))



# coding: utf-8
"""This module implements the JSON geometry and spatial reference objects 
   as returned by the REST API. The REST API supports 4 geometry types - 
   points, polylines, polygons and envelopes."""

import json

from .projections import projected, geographic

try:
    long, unicode, basestring
except NameError:
    long, unicode, basestring = int, str, str

def pointlist(points, sr):
    """Convert a list of the form [[x, y] ...] to a list of Point instances
       with the given x, y coordinates."""
    assert all(isinstance(pt, Point) or len(pt) == 2 
               for pt in points), "Point(s) not in [x, y] form"
    return [coord if isinstance(coord, Point) 
                  else Point(coord[0], coord[1], sr)
            for coord in points]

def listofpointlist(ptlist, sr):
    """Convert a list of the form [[[x, y] ...] ...] to a list of lists of 
       Point instances with the given x, y coordinates."""
    return [pointlist(listofpoints, sr) for listofpoints in ptlist]

class Geometry(object):
    """Represents an abstract base for json-represented geometries on
       the ArcGIS Server REST API. Please refer to 
       L{Point<arcrest.geometry.Point>}, 
       L{Multipoint<arcrest.geometry.Multipoint>},
       L{Polygon<arcrest.geometry.Polygon>},
       L{Polyline<arcrest.geometry.Polyline>} and 
       L{Envelope<arcrest.geometry.Envelope>} in this module for more
       information on geometry types. Calling the str() operator on any
       geometry subclass will return a WKT string for the geometry."""
    def __init__(self):
        raise NotImplementedError("Cannot instantiate abstract geometry type")
    def __len__(self):
        raise NotImplementedError("Length not implemented for %r" % 
                                   self.__class__.__name__)
    @property
    def __geo_interface__(self):
        raise NotImplementedError("Unimplemented conversion to GeoJSON")
    @property
    def _json_struct_without_sr(self):
        return self._json_struct
    @property
    def _json_struct(self):
        raise NotImplementedError("Unimplemented conversion to JSON")
    @property
    def _json_struct_without_sr(self):
        raise NotImplementedError("Unimplemented conversion to JSON")
    @property
    def _json_struct_for_featureset(self):
        return { 'geometry': self._json_struct_without_sr,
                 'attributes': getattr(self, 'attributes', {})}
    def __str__(self):
        return json.dumps(self._json_struct)
    @classmethod
    def fromJson(cls, struct):
        raise NotImplementedError("Unimplemented convert from JSON")
    @classmethod
    def fromGeoJson(cls, struct):
        raise NotImplementedError("Unimplemented convert from GeoJSON")

class NullGeometry(Geometry):
    "Represents a null or empty geometry."
    def __init__(self, struct=None):
        pass
    @property
    def __geo_interface__(self):
        return None
    @property
    def _json_struct(self):
        return None
    def __repr__(self):
        return "NULL GEOMETRY"

class SpatialReference(Geometry):
    """The REST API only supports spatial references that have well-known 
       IDs associated with them. Given this constraint, a spatial reference
       object only contains one field - wkid (i.e. the well-known ID of the
       spatial reference).

       For a list of valid WKID values, see projected and 
       projections.graphic in this package."""
    def __init__(self, wkid):
        """Create a new Spatial Reference.
           
                >>> import arcrest
                >>> mysr = arcrest.geometry.SpatialReference(4326)
                >>> mysr.name
                'GCS_WGS_1984'
                >>> mysr.wkid
                4326
                >>> myothersr = arcrest.geometry.SpatialReference(mysr)
                >>> myothersr.wkid
                4326

           @param wkid: The Well-known ID of the target spatial reference
                        (or another instance of SpatialReference)
           """
        if isinstance(wkid, SpatialReference):
            wkid = wkid.wkid
        elif isinstance(wkid, dict):
            wkid = wkid['wkid']
        elif hasattr(projected, str(wkid)):
            wkid = getattr(projected, str(wkid))
        elif hasattr(geographic, str(wkid)):
            wkid = getattr(geographic, str(wkid))
        elif wkid is None:
            self.wkid = None
            return
        if wkid is not None:
            wkid = int(wkid)
        self.wkid = wkid
    def __repr__(self):
        return "<Spatial Reference %r>" % self.wkid
    def __len__(self):
        if self.wkid is None:
            return 0
        else:
            return 1
    @property
    def _json_struct(self):
        return {'wkid': self.wkid}
    def __eq__(self, other):
        if isinstance(other, SpatialReference):
            return self.wkid == other.wkid
        return self.wkid == other
    @property
    def name():
        "Get/view the name for the well known ID of a Projection"
        if self.wkid in projected:
            return projected[self.wkid]
        elif self.wkid in geographic:
            return geographic[self.wkid]
        else:
            raise KeyError("Not a known WKID.")
    @name.setter
    def name(self, wkid):
        if hasattr(projected, wkid):
            self.wkid = getattr(projected, wkid)
        elif hasattr(geographic, wkid):
            self.wkid = getattr(geographic, wkid)
        else:
            raise KeyError("Not a known projection name.")
    @classmethod
    def fromJson(cls, struct):
        return cls(int(struct['wkid']))

class Point(Geometry):
    """A point contains x and y fields along with a spatialReference field."""
    __geometry_type__ = "esriGeometryPoint"
    def __init__(self, x, y, spatialReference=None):
        """
        @param x: The X coordinate of the Point
        @param y: The Y coordinate of the Point
        @param spatialReference: The spatial reference, either as an instance
               of L{SpatialReference<arcpy.geometry.SpatialReference>} or the
               WKID of a spatial reference. If left as None, the point will not
               have a spatial reference.
               
                    >>> mysr = arcrest.geometry.SpatialReference(4326)
                    >>> arcrest.geometry.Point(10, 10, mysr)
                    POINT(10.00000 10.00000)
               """
        if not isinstance(spatialReference, SpatialReference):
            spatialReference = SpatialReference(spatialReference)
        self.x, self.y, self.spatialReference = \
            float(x), float(y), spatialReference
    def __repr__(self):
        return "POINT(%s %s)" %(("%0.5f" % self.x)
                                           if isinstance(self.x, float)
                                           else str(self.x),
                                ("%0.5f" % self.y)
                                           if isinstance(self.y, float)
                                           else str(self.y))
    def __len__(self):
        return 2
    def __iter__(self):
        yield self.x
        yield self.y
    def __getitem__(self, index):
        return [self.x, self.y][index]
    @property
    def __geo_interface__(self):
        retval = {
            'type': 'Point',
            'coordinates': [self.x, self.y]
        }
        if hasattr(self, 'attributes'):
            retval['properties'] = self.attributes
        if self.spatialReference:
            retval['@esri.sr'] = self.spatialReference._json_struct
        return retval
    @property
    def _json_struct_without_sr(self):
        return {'x': self.x,
                'y': self.y}
    @property
    def _json_struct(self):
        return {'x': self.x,
                'y': self.y,
                'spatialReference': None \
                                    if self.spatialReference is None \
                                    else self.spatialReference._json_struct}
    @classmethod
    def fromJson(cls, struct):
        if isinstance(struct, (list, tuple)) and len(struct) == 2:
            return cls(*struct)
        else:
            return cls(**struct)
    @classmethod
    def fromGeoJson(cls, struct):
        (x, y) = struct['coordinates']
        return [cls(x, y)]

class Polyline(Geometry):
    """A polyline contains an array of paths and a spatialReference. Each 
       path is represented as an array of points. And each point in the path is
       represented as a 2-element array. The 0-index is the x-coordinate and
       the 1-index is the y-coordinate."""
    __geometry_type__ = "esriGeometryPolyline"
    def __init__(self, paths=[], spatialReference=None):
        """
        @param paths: A list of lists of points. Actual acceptable values are
                      fairly permissive, allowing for any iterable item 
                      (list, tuple, generator) containing any number of 
                      iterables containing instances of 
                      L{Point<arcrest.geometry.Point>} or lists/tuples of
                      exactly two items, representing a coordinate pair.

        @param spatialReference: A spatial reference passed in as an instance
                                 of SpatialReference or the WKID of a spatial
                                 reference. If this is not set, the Polyline
                                 will not have a spatial reference and if the
                                 polyline is used in a context where a spatial
                                 reference is required, it will attempt to
                                 guess the spatial reference from the set
                                 spatial reference of a constitutent point.
        """
        if not isinstance(spatialReference, SpatialReference):
            spatialReference = SpatialReference(spatialReference)
        self.spatialReference = spatialReference
        self.paths = listofpointlist(paths, spatialReference)
    def __repr__(self):
        return "MULTILINESTRING(%s)" % " ".join(
                                        "(%s)"%"".join(
                                           ",".join(
                                                " ".join("%0.5f"%x 
                                                       if isinstance(x, float)
                                                       else str(x) for x in pt)
                                            for pt in path)) 
                                        for path in self._json_paths)
    def __len__(self):
        return len(self.paths)
    @property
    def __geo_interface__(self):
        retval = {
            'type': 'MultiLineString',
            'coordinates': self._json_paths
        }
        if hasattr(self, 'attributes'):
            retval['properties'] = self.attributes
        if self.spatialReference:
            retval['@esri.sr'] = self.spatialReference._json_struct
        return retval
    @property
    def _json_paths(self):
        def fixpath(somepath):
            for pt in somepath:
                if isinstance(pt, Point):
                    assert pt.spatialReference == None or\
                        pt.spatialReference == self.spatialReference, \
                        "Point is not in same spatial reference as Polyline"\
                        "(%r, %r)" % (pt.spatialReference, 
                                      self.spatialReference)
                    yield [pt.x, pt.y]
                else:
                    yield list(pt)
        return [list(fixpath(path)) for path in self.paths]
    @property
    def _json_struct_without_sr(self):
        return {'paths': self._json_paths}
    @property
    def _json_struct(self):
        return {'paths': self._json_paths,
                'spatialReference': self.spatialReference._json_struct}
    @classmethod
    def fromJson(cls, struct):
        return cls(**struct)
    @classmethod
    def fromGeoJson(cls, struct):
        if struct['type'] == "LineString":
            return [cls([struct['coordinates']])]
        elif struct['type'] == "MultiLineString":
            return [cls(struct['coordinates'])]
    @classmethod
    def fromCompressedGeometry(cls, compressedstring, attributes=None):
        result = []
        import re
        ints = [(-1 if number[0] == '-' else 1) * int(number[1:], 32)
                   for number in re.findall("([+-][a-v0-9]*)",
                                            compressedstring)]
        multiplier = float(ints.pop(0))
        oldx, oldy = 0, 0
        while ints:
            x, y = ints[:2]
            ints = ints[2:]
            x += oldx
            y += oldy
            result.append(Point(x/multiplier, y/multiplier))
            oldx, oldy = x, y
        retval = cls([result])
        if attributes:
            retval.attributes = attributes
        return retval
    def asCompressedGeometry(self, multiplier=55000):
        def base32(num):
            sign = "+" if num >= 0 else "-"
            if num < 0:
                num = -num
            digits = "0123456789abcdefghijklmnopqrstuv"
            nums = []
            while num:
                nums.append(num & 0x1F)
                num = num >> 5
            return sign + ''.join(digits[x] for x in reversed(nums))
        def compressedstring():
            yield base32(multiplier)
            oldx, oldy = 0, 0
            for ints in self._json_paths:
                for (x, y) in ints:
                    yield base32(int((x - oldx)*multiplier))
                    yield base32(int((y - oldy)*multiplier))
                    oldx, oldy = x, y
        return ''.join(compressedstring())

class Polygon(Geometry):
    """A polygon contains an array of rings and a spatialReference. Each ring 
       is represented as an array of points. The first point of each ring is
       always the same as the last point. And each point in the ring is 
       represented as a 2-element array. The 0-index is the x-coordinate and
       the 1-index is the y-coordinate."""
    __geometry_type__ = "esriGeometryPolygon"
    def __init__(self, rings=[], spatialReference=None):
        """
        @param rings: A list of lists of points. Actual acceptable values are
                      fairly permissive, allowing for any iterable item 
                      (list, tuple, generator) containing any number of 
                      iterables containing instances of 
                      L{Point<arcrest.geometry.Point>} or lists/tuples of
                      exactly two items, representing a coordinate pair. Note
                      that the first and last point of each ring I{must} be
                      the same.

        @param spatialReference: A spatial reference passed in as an instance
                                 of SpatialReference or the WKID of a spatial
                                 reference. If this is not set, the Polyline
                                 will not have a spatial reference and if the
                                 polyline is used in a context where a spatial
                                 reference is required, it will attempt to
                                 guess the spatial reference from the set
                                 spatial reference of a constitutent point.
        """
        if not isinstance(spatialReference, SpatialReference):
            spatialReference = SpatialReference(spatialReference)
        self.spatialReference = spatialReference
        self.rings = listofpointlist(rings, spatialReference)
    def __repr__(self):
        return "POLYGON(%s)" % " ".join(
                                        "(%s)"%"".join(
                                           ",".join(
                                                " ".join("%0.5f"%x 
                                                      if isinstance(x, float)
                                                      else str(x) for x in pt)
                                            for pt in ring)) 
                                        for ring in self._json_rings)
    def __len__(self):
        return len(self.rings)
    @property
    def __geo_interface__(self):
        retval = {
            'type': 'Polygon',
            'coordinates': self._json_rings
        }
        if hasattr(self, 'attributes'):
            retval['properties'] = self.attributes
        if self.spatialReference:
            retval['@esri.sr'] = self.spatialReference._json_struct
        return retval
    def contains(self, pt):
        "Tests if the provided point is in the polygon."
        if isinstance(pt, Point):
            ptx, pty = pt.x, pt.y
            assert (self.spatialReference is None or \
                    self.spatialReference.wkid is None) or \
                    (pt.spatialReference is None or \
                     pt.spatialReference.wkid is None) or \
                   self.spatialReference == pt.spatialReference, \
                   "Spatial references do not match."
        else:
            ptx, pty = pt
        in_shape = False
        # Ported nearly line-for-line from the Javascript API
        for ring in self._json_rings:
            for idx in range(len(ring)):
                idxp1 = idx + 1
                if idxp1 >= len(ring):
                    idxp1 -= len(ring)
                pi, pj = ring[idx], ring[idxp1]
                # Divide-by-zero checks
                if (pi[1] == pj[1]) and pty >= min((pi[1], pj[1])):
                    if ptx >= max((pi[0], pj[0])):
                        in_shape = not in_shape
                elif (pi[0] == pj[0]) and pty >= min((pi[0], pj[0])):
                    if ptx >= max((pi[1], pj[1])):
                        in_shape = not in_shape
                elif (((pi[1] < pty and pj[1] >= pty) or 
                     (pj[1] < pty and pi[1] >= pty)) and 
                    (pi[0] + (pty - pi[1]) / 
                     (pj[1] - pi[1]) * (pj[0] - pi[0]) < ptx)):
                    in_shape = not in_shape
        return in_shape
    def __contains__(self, pt):
        return self.contains(pt)
    @property
    def _json_rings(self):
        def fixring(somering):
            for pt in somering:
                if isinstance(pt, Point):
                    assert pt.spatialReference == None or \
                        pt.spatialReference == self.spatialReference, \
                        "Point is not in the same spatial reference as Polygon"
                    yield [pt.x, pt.y]
                else:
                    yield list(pt)
        return [list(fixring(ring)) for ring in self.rings]
    @property
    def _json_struct_without_sr(self):
        return {'rings': self._json_rings}
    @property
    def _json_struct(self):
        return {'rings': self._json_rings,
                'spatialReference': self.spatialReference._json_struct}
    @classmethod
    def fromJson(cls, struct):
        return cls(**struct)
    @classmethod
    def fromGeoJson(cls, struct):
        if struct['type'] == "MultiPolygon":
            return [cls(x) for x in struct['coordinates']]
        else:
            return [cls(struct['coordinates'])]

class Multipoint(Geometry):
    """A multipoint contains an array of points and a spatialReference. Each
       point is represented as a 2-element array. The 0-index is the
       x-coordinate and the 1-index is the y-coordinate."""
    def __init__(self, points=[], spatialReference=None):
        if not isinstance(spatialReference, SpatialReference):
            spatialReference = SpatialReference(spatialReference)
        self.spatialReference = spatialReference
        self.points = pointlist(points, spatialReference)
    def __repr__(self):
        return "MULTIPOINT(%s)" % ",".join("%0.5f %0.5f" % tuple(map(float,
                                                                     pt))
                                           for pt in self._json_points)
    def __len__(self):
        return len(self.points)
    @property
    def __geo_interface__(self):
        retval = {
            'type': 'MultiPoint',
            'coordinates': self._json_points
        }
        if hasattr(self, 'attributes'):
            retval['properties'] = self.attributes
        if self.spatialReference:
            retval['@esri.sr'] = self.spatialReference._json_struct
        return retval
    @property
    def _json_points(self):
        def fixpoint(somepointarray):
            for pt in somepointarray:
                if isinstance(pt, Point):
                    assert pt.spatialReference == None or \
                        pt.spatialReference == self.spatialReference, \
                        "Point is not in same spatial reference as Multipoint"
                    yield [pt.x, pt.y]
                else:
                    yield list(pt)
        return list(fixpoint(self.points))
    @property
    def _json_struct_without_sr(self):
        return {'points': self._json_points}
    @property
    def _json_struct(self):
        return {'points': self._json_points,
                'spatialReference': self.spatialReference._json_struct}
    @classmethod
    def fromJson(cls, struct):
        return cls(**struct)
    @classmethod
    def fromGeoJson(cls, struct):
        return [cls(struct['coordinates'])]

class Envelope(Geometry):
    """An envelope contains the corner points of an extent and is represented
       by xmin, ymin, xmax, and ymax, along with a spatialReference."""
    __geometry_type__ = "esriGeometryEnvelope"
    def __init__(self, xmin, ymin, xmax, ymax,
                 spatialReference=None):
        if not isinstance(spatialReference, SpatialReference):
            spatialReference = SpatialReference(spatialReference)
        self.spatialReference = spatialReference
        self.xmin, self.ymin, self.xmax, self.ymax = \
            float(xmin), float(ymin), float(xmax), float(ymax)
    def __contains__(self, pt):
        if isinstance(pt, Point):
            assert pt.spatialReference is None or \
                   self.spatialReference is None or\
                   pt.spatialReference == self.spatialReference, \
                   "Incompatible spatial reference for point"
            x, y = pt.x, pt.y
        else:
            x, y = pt[:2]
        return (self.xmax >= x >= self.xmin) and (self.ymax >= y >= self.ymin)
    def __bool__(self):
        return bool(self.wkid is not None)
    @property
    def __geo_interface__(self):
        retval = {
            'type': 'Box',
            'coordinates': [[self.xmin, self.ymin], [self.xmax, self.ymax]]
        }
        if hasattr(self, 'attributes'):
            retval['properties'] = self.attributes
        if self.spatialReference:
            retval['@esri.sr'] = self.spatialReference._json_struct
        return retval
    @property
    def top(self):
        return Point(self.xmin, self.ymin, self.spatialReference)
    @property
    def bottom(self):
        return Point(self.xmax, self.ymax, self.spatialReference)
    @property
    def _json_struct_without_sr(self):
        return {'xmin': self.xmin, 
                'ymin': self.ymin,
                'xmax': self.xmax,
                'ymax': self.ymax}
    @property
    def _json_struct(self):
        return {'xmin': self.xmin,
                'ymin': self.ymin,
                'xmax': self.xmax,
                'ymax': self.ymax,
                'spatialReference': self.spatialReference._json_struct}
    @property
    def bbox(self):
        "Return the envelope as a Bound Box string compatible with (bb) params"
        return ",".join(str(attr) for attr in 
                            (self.xmin, self.ymin, self.xmax, self.ymax))
    @classmethod
    def fromJson(cls, struct):
        return cls(**struct)
    @classmethod
    def fromGeoJson(cls, struct):
        ((xmin, ymin), (xmax, ymax)) = struct['coordinates']
        return [cls(xmin, ymin, xmax, ymax)]

def fromJson(struct, attributes=None):
    "Convert a JSON struct to a Geometry based on its structure"
    if isinstance(struct, basestring):
        struct = json.loads(struct)
    indicative_attributes = {
        'x': Point,
        'wkid': SpatialReference,
        'paths': Polyline,
        'rings': Polygon,
        'points': Multipoint,
        'xmin': Envelope
    }
    # bbox string
    if isinstance(struct, basestring) and len(struct.split(',')) == 4:
        return Envelope(*map(float, struct.split(',')))
    # Look for telltale attributes in the dict
    if isinstance(struct, dict):
        for key, cls in indicative_attributes.items():
            if key in struct:
                ret = cls.fromJson(dict((str(key), value)
                                   for (key, value) in struct.items()))
                if attributes:
                    ret.attributes = dict((str(key.lower()), val) 
                                           for (key, val)
                                           in attributes.items())
                return ret
    raise ValueError("Unconvertible to geometry")

def fromGeoJson(struct, attributes=None):
    "Convert a GeoJSON-like struct to a Geometry based on its structure"
    if isinstance(struct, basestring):
        struct = json.loads(struct)
    type_map = {
        'Point': Point,
        'MultiLineString': Polyline,
        'LineString': Polyline,
        'Polygon': Polygon,
        'MultiPolygon': Polygon,
        'MultiPoint': Multipoint,
        'Box': Envelope
    }
    if struct['type'] == "Feature":
        return fromGeoJson(struct, struct.get('properties', None))
    elif struct['type'] == "FeatureCollection":
        sr = None
        if 'crs' in struct:
            sr = SpatialReference(struct['crs']['properties']['code'])
            members = map(fromGeoJson, struct['members'])
            for member in members:
                member.spatialReference = sr
            return members
        else:
            return map(fromGeoJson, struct['members'])
    elif struct['type'] in type_map and hasattr(type_map[struct['type']], 
                                              'fromGeoJson'):
        instances = type_map[struct['type']].fromGeoJson(struct)
        i = []
        assert instances is not None, "GeoJson conversion returned a Null geom"
        for instance in instances:
            if 'properties' in struct:
                instance.attributes = struct['properties'].copy()
                if '@esri.sr' in instance.attributes:
                    instance.spatialReference = SpatialReference.fromJson(
                                               instance.attributes['@esri.sr'])
                    del instance.attributes['@esri.sr']
            if attributes:
                if not hasattr(instance, 'attributes'):
                    instance.attributes = {}
                for k, v in attributes.items():
                    instance.attributes[k] = v
            i.append(instance)
        if i:
            if len(i) > 1:
                return i
            return i[0]
    raise ValueError("Unconvertible to geometry")
